@@ -354,8 +354,8 @@ gst_cairo_render_base_init (gpointer g_class)
   gst_element_class_set_details_simple (ec, "Cairo encoder",
       "Codec/Encoder", "Encodes streams using Cairo",
       "Lutz Mueller <lutz@topfrose.de>");
-  gst_element_class_add_static_pad_template (ec, &t_snk);
-  gst_element_class_add_static_pad_template (ec, &t_src);
+  gst_element_class_add_pad_template (ec, gst_static_pad_template_get (&t_snk));
+  gst_element_class_add_pad_template (ec, gst_static_pad_template_get (&t_src));
 }
 
 static void
